import copy
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import altair as alt
import click
import hdbscan
import numpy as np
import pandas as pd
import streamlit as st
from sklearn.cluster import KMeans
from sklearn.neighbors import VALID_METRICS as SKLEARN_DISTANCE_METRICS
from umap import UMAP
from umap.distances import named_distances

from gobbli.interactive.util import (
    get_label_indices,
    load_data,
    safe_sample,
    st_sample_data,
    st_select_model_checkpoint,
    st_select_untrained_model,
)
from gobbli.io import EmbedInput, EmbedPooling
from gobbli.model import FastText
from gobbli.model.mixin import EmbedMixin
from gobbli.util import TokenizeMethod, tokenize, truncate_text

DEFAULT_EMBED_BATCH_SIZE = EmbedInput.embed_batch_size

# For performance, don't let streamlit try to hash the tokens.  It takes forever
@st.cache(allow_output_mutation=True)
def get_tokens(
    texts: List[str], tokenize_method: TokenizeMethod, vocab_size: int
) -> List[List[str]]:
    try:
        return tokenize(tokenize_method, texts, vocab_size=vocab_size)
    except RuntimeError as e:
        str_e = str(e)
        if "vocab_size()" in str_e and "pieces_size()" in str_e:
            st.error(
                "SentencePiece requires your texts to have at least as many different tokens "
                "as its vocabulary size.  Try a smaller vocabulary size."
            )
            return
        else:
            raise


def _show_example_documents(
    texts: List[str], labels: Optional[List[str]], truncate_len: int
):
    df = pd.DataFrame({"Document": [truncate_text(t, truncate_len) for t in texts]})
    if labels is not None:
        df["Label"] = labels
    st.table(df)


def show_example_documents(
    texts: List[str],
    labels: List[str],
    filter_label: Optional[str],
    example_truncate_len: int,
    example_num_docs: int,
):
    st.header("Example Documents")

    # If we're filtered to a specific label,
    # just show it once at the top -- otherwise, show the label
    # with each example
    if filter_label is not None:
        st.subheader(f"Label: {filter_label}")
        example_labels = None
    else:
        example_labels = labels

    example_indices = safe_sample(range(len(texts)), example_num_docs)
    _show_example_documents(
        [texts[i] for i in example_indices],
        [example_labels[i] for i in example_indices]
        if example_labels is not None
        else None,
        example_truncate_len,
    )


def get_document_lengths(tokens: List[List[str]]) -> List[int]:
    return [len(d) for d in tokens]


def _collect_label_counts(labels: List[str]) -> pd.DataFrame:
    label_counts = (
        pd.Series(labels).value_counts(normalize=True).to_frame(name="Proportion")
    )
    label_counts.index.name = "Label"
    return label_counts.reset_index()


def show_label_distribution(
    sample_labels: List[str], all_labels: Optional[List[str]] = None
):
    if sample_labels is not None:
        st.header("Label Distribution")
        label_counts = _collect_label_counts(sample_labels)

        if all_labels is None:
            label_chart = (
                alt.Chart(label_counts, height=500, width=700)
                .mark_bar()
                .encode(
                    alt.X("Label", type="nominal"),
                    alt.Y("Proportion", type="quantitative"),
                )
            )
        else:
            label_counts["Label Set"] = "Sample"
            all_label_counts = _collect_label_counts(all_labels)
            all_label_counts["Label Set"] = "All Documents"
            label_counts = pd.concat([label_counts, all_label_counts])

            label_chart = (
                alt.Chart(label_counts, width=100)
                .mark_bar()
                .encode(
                    alt.X(
                        "Label Set",
                        type="nominal",
                        title=None,
                        sort=["Sample", "All Documents"],
                    ),
                    alt.Y("Proportion", type="quantitative"),
                    alt.Column(
                        "Label", type="nominal", header=alt.Header(labelAngle=0)
                    ),
                    alt.Color("Label Set", type="nominal", legend=None),
                )
            )

        st.altair_chart(label_chart)


def show_document_length_distribution(tokens: List[List[str]]):
    st.header("Document Length Distribution")
    document_lengths = get_document_lengths(tokens)
    doc_lengths = pd.DataFrame({"Token Count": document_lengths})
    doc_length_chart = (
        alt.Chart(doc_lengths, height=500, width=700)
        .mark_bar()
        .encode(
            alt.X("Token Count", bin=alt.Bin(maxbins=30)),
            alt.Y("count()", type="quantitative"),
        )
    )

    st.altair_chart(doc_length_chart)


@st.cache(show_spinner=True)
def get_topics(
    tokens: List[List[str]],
    num_topics: int = 10,
    train_chunksize: int = 2000,
    train_passes: int = 3,
    train_iterations: int = 100,
    do_bigrams: bool = True,
    bigram_min_count: int = 20,
    min_frequency: int = 20,
    max_proportion: float = 0.5,
) -> List[Tuple[List[Tuple[float, str]], float]]:
    """
    Calculate topics and topic assignments for the given tokens.

    Args:
      tokens: Tokenized documents.
      num_topics: Number of topics to pull out of the trained model.
      train_chunksize: Number of documents to process at a time during training.
      train_passes: Number of full passes to make over the dataset.
      train_iterations: Maximum number of iterations through the corpus.
      do_bigrams: If True, use bigram features in addition to unigram tokens.
      bigram_min_count: Minimum number of times 2 tokens need to be colocated to
        be considered a bigram.
      min_frequency: Minimum frequency across all documents for a token to be included
        in the vocabulary.
      max_proportion: Maximum proportion of documents containing a token for it to be
        included in the vocabulary.

    Returns:
      2-tuple: The calculated topic structure as returned by gensim and a list containing,
      for each document, the predicted topic ID for that document.
    """
    try:
        import gensim
    except ImportError:
        raise ImportError("Topic modeling requires gensim to be installed.")

    # We need to be able to mutate this list
    gensim_tokens = copy.deepcopy(tokens)

    if do_bigrams:
        bigram = gensim.models.Phrases(gensim_tokens, min_count=bigram_min_count)
        for ndx in range(len(gensim_tokens)):
            for token in bigram[gensim_tokens[ndx]]:
                if "_" in token:
                    # Token is a bigram -- add to document.
                    gensim_tokens[ndx].append(token)

    dictionary = gensim.corpora.Dictionary(tokens)
    dictionary.filter_extremes(no_below=min_frequency, no_above=max_proportion)
    corpus = [dictionary.doc2bow(doc) for doc in gensim_tokens]

    _ = dictionary[0]  # "Load" the dictionary
    id2word = dictionary.id2token

    model = gensim.models.LdaModel(
        corpus=corpus,
        id2word=id2word,
        chunksize=train_chunksize,
        alpha="auto",
        eta="auto",
        iterations=train_iterations,
        num_topics=num_topics,
        passes=train_passes,
        eval_every=None,
    )

    topic_assignments = []
    for doc_bow in corpus:
        topic_probs = model.get_document_topics(doc_bow)
        assigned_topic, _ = max(topic_probs, key=lambda t: t[1])
        topic_assignments.append(assigned_topic)

    return model.top_topics(corpus), topic_assignments


def corr_df_to_heatmap_df(
    corr_df: pd.DataFrame, index_col_name: str, columns_col_name: str
) -> pd.DataFrame:
    """
    Convert a dataframe consisting of a correlation matrix (index values x column values)
    into a long dataframe which can drive an Altair heatmap.
    """
    heatmap_df = corr_df.copy()
    heatmap_df.index.name = index_col_name
    heatmap_df = heatmap_df.reset_index()
    return heatmap_df.melt(
        id_vars=index_col_name, var_name=columns_col_name, value_name="Correlation"
    )


def st_heatmap(
    heatmap_df: pd.DataFrame, x_col_name: str, y_col_name: str, color_col_name: str
):
    heatmap = (
        alt.Chart(heatmap_df, height=700, width=700)
        .mark_rect()
        .encode(alt.X(x_col_name), alt.Y(y_col_name), alt.Color(color_col_name))
    )
    st.altair_chart(heatmap)


def show_topic_model(
    tokens: List[List[str]],
    labels: Optional[List[str]],
    filter_label: Optional[str],
    **model_kwargs,
):
    topics, topic_assignments = get_topics(tokens, **model_kwargs)

    if filter_label is not None:
        st.subheader(f"Label: {filter_label}")

    topic_df_data = []
    for i, (topic, coherence) in enumerate(topics):
        topic_data = {"Coherence": coherence}

        for j, (probability, word) in enumerate(topic):
            topic_data[j + 1] = f"{word} ({probability:.4f})"

        topic_df_data.append(topic_data)

    topic_df = pd.DataFrame(topic_df_data).T
    topic_df.columns = [f"Topic {i+1}" for i in range(len(topic_df.columns))]

    # This is enough to show the full df, but streamlit renders it with a scrollbar still,
    # so there's still a little bit of scroll
    st.dataframe(topic_df, height=750)

    # Combine all topics and labels together to get topic-label, topic-topic, and
    # label-label correlations
    topic_onehot = pd.get_dummies([str(t) for t in topic_assignments])
    label_onehot = pd.get_dummies(labels)
    heatmap_data = pd.concat(
        # Convert topic IDs to strings so Altair doesn't try to make them numeric
        [topic_onehot, label_onehot],
        axis=1,
    )
    heatmap_corr = heatmap_data.corr()

    # Pull out topic-topic correlations and display
    all_topics = topic_onehot.columns
    topic_topic_matrix = heatmap_corr.loc[all_topics, all_topics]

    st.subheader("Topic - Topic Correlation")
    st_heatmap(
        corr_df_to_heatmap_df(topic_topic_matrix, "Topic 1", "Topic 2"),
        "Topic 1",
        "Topic 2",
        "Correlation",
    )

    if labels is not None:
        # Pull out topic-label correlations and display
        all_labels = label_onehot.columns
        topic_label_matrix = heatmap_corr.loc[all_topics, all_labels]

        st.subheader("Topic - Label Correlation")
        st_heatmap(
            corr_df_to_heatmap_df(topic_label_matrix, "Topic", "Label"),
            "Topic",
            "Label",
            "Correlation",
        )


@st.cache(show_spinner=True)
def get_embeddings(
    model_cls: Any,
    model_kwargs: Dict[str, Any],
    texts: List[str],
    checkpoint_meta: Optional[Dict[str, Any]] = None,
    batch_size: int = DEFAULT_EMBED_BATCH_SIZE,
    pooling: EmbedPooling = EmbedPooling.MEAN,
) -> Tuple[List[np.ndarray], Optional[List[List[str]]]]:
    """
    Generate embeddings using the given model and optional trained checkpoint.

    Args:
      model_cls: Class of model to use.
      model_kwargs: Parameters to pass when initializing the model.
      texts: Texts to generate embeddings for.
      checkpoint_meta: Optional metadata for a trained checkpoint.  If passed, a trained model
        will be used.
      batch_size: Batch size to use when generating embeddings.
      pooling: Pooling method for combining embeddings across documents.

    Returns:
      A 2-tuple: First element is the generated embeddings - a numpy array for each passed
      document.  If pooling method is NONE, the second element is the corresponding nested list of
      tokens; otherwise, the second element is None.
    """
    embed_input = EmbedInput(
        X=texts,
        checkpoint=None
        if checkpoint_meta is None
        else Path(checkpoint_meta["checkpoint"]),
        embed_batch_size=batch_size,
        pooling=pooling,
    )
    model = model_cls(**model_kwargs)
    model.build()
    embed_output = model.embed(embed_input)
    if pooling == EmbedPooling.NONE:
        return embed_output.X_embedded, embed_output.embed_tokens
    else:
        return embed_output.X_embedded, None


def show_embeddings(
    model_cls: Any,
    model_kwargs: Dict[str, Any],
    texts: List[str],
    labels: Optional[List[str]],
    checkpoint_meta: Optional[Dict[str, Any]] = None,
    batch_size: int = DEFAULT_EMBED_BATCH_SIZE,
    umap_seed: int = 1,
    umap_n_neighbors: int = 15,
    umap_metric: str = "euclidean",
    umap_min_dist: float = 0.1,
    cluster_when: str = "before",
    clusterer: Optional[Any] = None,
    show_vocab_overlap: bool = False,
):
    if cluster_when not in ("before", "after"):
        raise ValueError(f"Unexpected cluster_when value: '{cluster_when}'")

    embeddings, _ = get_embeddings(
        model_cls,
        model_kwargs,
        texts,
        checkpoint_meta=checkpoint_meta,
        batch_size=batch_size,
    )
    X_embedded = pd.DataFrame(embeddings)

    umap = UMAP(
        n_neighbors=umap_n_neighbors,
        metric=umap_metric,
        min_dist=umap_min_dist,
        random_state=umap_seed,
    )

    clusters = None
    if clusterer is not None and cluster_when == "before":
        clusterer.fit(X_embedded)
        clusters = clusterer.labels_

    umap_data = umap.fit_transform(X_embedded)

    if clusterer is not None and cluster_when == "after":
        clusterer.fit(umap_data)
        clusters = clusterer.labels_

    umap_df = pd.DataFrame(umap_data, columns=["UMAP Component 1", "UMAP Component 2"])
    tooltip_attrs = ["Text"]

    if labels is not None:
        tooltip_attrs.append("Label")
        umap_df["Label"] = labels

    if clusters is not None:
        color_attr = "Cluster"
        tooltip_attrs.append("Cluster")
        umap_df["Cluster"] = clusters
        umap_df["Cluster"] = umap_df["Cluster"].astype(str)
    elif labels is not None:
        color_attr = "Label"
    else:
        color_attr = None

    # NOTE: Altair (or the underlying charting library, vega-lite) will
    # truncate these texts before being displayed
    umap_df["Text"] = texts

    umap_chart = (
        alt.Chart(umap_df, height=700, width=700)
        .mark_circle(size=60)
        .encode(
            alt.X("UMAP Component 1", scale=alt.Scale(zero=False), axis=None),
            alt.Y("UMAP Component 2", scale=alt.Scale(zero=False), axis=None),
            tooltip=alt.Tooltip(tooltip_attrs),
        )
    )
    if color_attr is not None:
        umap_chart = umap_chart.encode(alt.Color(color_attr))

    st.altair_chart(umap_chart)

    if show_vocab_overlap:
        missing_token_counts = defaultdict(int)
        oov_count = 0
        total_count = 0
        try:
            embeddings, embed_tokens = get_embeddings(
                model_cls,
                model_kwargs,
                texts,
                checkpoint_meta=checkpoint_meta,
                batch_size=batch_size,
                pooling=EmbedPooling.NONE,
            )
        except ValueError:
            st.error(
                "This model doesn't support generating token-level embeddings, "
                "so the vocabulary overlap report can't be calculated."
            )
            return

        for doc_embedding, doc_tokens in zip(embeddings, embed_tokens):
            oov_embeddings = np.abs(doc_embedding).sum(axis=1) == 0
            oov_count += oov_embeddings.sum()
            total_count += len(doc_tokens)
            for oov_token_ndx in np.argwhere(oov_embeddings).ravel():
                missing_token_counts[doc_tokens[oov_token_ndx]] += 1

        st.subheader("Vocabulary Overlap")
        num_oov_tokens_show = 20
        st.markdown(
            f"""
            - Total number of tokens: {total_count:,}
            - Number of out-of-vocabulary tokens: {oov_count:,} ({(oov_count / total_count) * 100:.2f}%)
            """
        )
        if oov_count > 0:
            st.markdown(
                f"""
                - Top {num_oov_tokens_show} most frequent out-of-vocabulary tokens:
                """
            )
            oov_df = pd.DataFrame.from_dict(
                {
                    tok: count
                    for tok, count in sorted(
                        missing_token_counts.items(), key=lambda kv: -kv[1]
                    )
                },
                orient="index",
            ).tail(num_oov_tokens_show)
            oov_df.columns = ["Out-of-Vocabulary Token Count"]
            st.dataframe(oov_df)


@click.command()
@click.argument("data", type=str)
@click.option(
    "--n-rows",
    type=int,
    help="Number of rows to load from the data file.  If -1, load all rows.",
    default=-1,
    show_default=True,
)
@click.option(
    "--model-data-dir",
    type=str,
    help="Optional data directory for a model.  If given, the model checkpoints "
    "will become available to use for embedding generation, if the model supports it.",
    default=None,
)
@click.option(
    "--use-gpu/--use-cpu",
    default=False,
    help="Which device to run the model on, if any. Defaults to CPU.",
)
@click.option(
    "--nvidia-visible-devices",
    default="all",
    help="Which GPUs to make available to the container; ignored if running on CPU. "
    "If not 'all', should be a comma-separated string: ex. ``1,2``.",
    show_default=True,
)
def run(
    data: str,
    n_rows: int,
    model_data_dir: Optional[str],
    use_gpu: bool,
    nvidia_visible_devices: str,
):
    st.title(f"Exploring: {data}")
    texts, labels = load_data(data, None if n_rows == -1 else n_rows)
    if labels is not None:
        label_indices = get_label_indices(labels)

    #
    # Sidebar
    #

    st.sidebar.header("Filter")

    filter_label = None
    if labels is not None and st.sidebar.checkbox(
        "Filter By Label", key="sample_by_label"
    ):
        filter_label = st.sidebar.selectbox("Label", list(sorted(label_indices.keys())))

    if filter_label is None:
        filtered_texts = texts
        filtered_labels = labels
    else:
        filtered_texts = [texts[i] for i in label_indices[filter_label]]
        filtered_labels = [labels[i] for i in label_indices[filter_label]]

    sampled_texts, sampled_labels = st_sample_data(filtered_texts, filtered_labels)

    st.sidebar.header("Examples")
    example_truncate_len = st.sidebar.number_input(
        "Example Truncate Length", min_value=1, max_value=None, value=500
    )
    example_num_docs = st.sidebar.number_input(
        "Number of Example Documents", min_value=1, max_value=None, value=5
    )

    st.sidebar.header("Labels")
    show_full_label_distribution = st.sidebar.checkbox("Show Full Label Distribution")

    st.sidebar.header("Tokenization")
    tokenize_method = TokenizeMethod[
        st.sidebar.selectbox("Method", tuple(tm.name for tm in TokenizeMethod))
    ]
    vocab_size_input = st.sidebar.empty()
    vocab_size = None
    if tokenize_method == TokenizeMethod.SENTENCEPIECE:
        vocab_size = vocab_size_input.number_input(
            "Vocabulary Size", min_value=1, max_value=None, value=20000
        )

    st.sidebar.header("Topic Model")
    run_topic_model = False
    if st.sidebar.checkbox("Enable Topic Model"):
        run_topic_model = st.sidebar.button("Train Topic Model")
        num_topics = st.sidebar.number_input(
            "Number of Topics", min_value=2, max_value=None, value=10
        )
        train_chunksize = st.sidebar.number_input(
            "Training Chunk Size", min_value=1, max_value=None, value=2000
        )
        train_passes = st.sidebar.number_input(
            "Number of Training Passes", min_value=1, max_value=None, value=3
        )
        train_iterations = st.sidebar.number_input(
            "Number of Training Iterations", min_value=1, max_value=None, value=100
        )
        do_bigrams = st.sidebar.checkbox("Use Bigrams?", value=True)
        if do_bigrams:
            bigram_min_count = st.sidebar.number_input(
                "Minimum Count for Bigrams", min_value=1, max_value=None, value=20
            )
        min_frequency = st.sidebar.number_input(
            "Minimum Vocabulary Frequency", min_value=1, max_value=None, value=20
        )
        max_proportion = st.sidebar.number_input(
            "Maximum Vocabulary Proportion", min_value=0.0, max_value=1.0, value=0.5
        )

    st.sidebar.header("Embeddings")
    run_embeddings = False
    embeddings_error = False
    embed_batch_size = DEFAULT_EMBED_BATCH_SIZE
    if st.sidebar.checkbox("Enable Embeddings"):
        run_embeddings = st.sidebar.button("Generate Embeddings")
        model_type_options = ["Untrained"]
        if model_data_dir is not None:
            model_type_options.insert(0, "Trained")
        else:
            st.sidebar.markdown(
                "**Note**: Using a trained model requires passing a model to this app "
                "via the --model-data-dir command line argument."
            )
        model_type = st.sidebar.selectbox("Model Type", model_type_options)

        if model_type == "Trained" and model_data_dir is not None:
            result = st_select_model_checkpoint(
                Path(model_data_dir), use_gpu, nvidia_visible_devices
            )
            if result is None:
                embeddings_error = True
                st.sidebar.error("Error selecting model checkpoint.")
            else:
                model_cls, model_kwargs, checkpoint_meta = result
        else:
            checkpoint_meta = None
            model_result = st_select_untrained_model(
                use_gpu,
                nvidia_visible_devices,
                # We want only models that support embeddings and aren't
                # fastText, since fastText requires training before generating embeddings
                predicate=lambda m: issubclass(m, EmbedMixin) and m is not FastText,
            )

            if model_result is None:
                run_embeddings = False
                embeddings_error = True
            else:
                model_cls, model_kwargs = model_result

        embed_batch_size = st.sidebar.number_input(
            "Batch Size",
            min_value=1,
            max_value=len(sampled_texts),
            value=min(len(sampled_texts), DEFAULT_EMBED_BATCH_SIZE),
        )

        umap_seed = st.sidebar.number_input(
            "UMAP Random Seed", min_value=1, max_value=None, value=1
        )
        umap_n_neighbors = st.sidebar.number_input(
            "UMAP Number of Neigbors", min_value=1, max_value=None, value=15
        )
        umap_metric = st.sidebar.selectbox(
            "UMAP Distance Metric", list(named_distances.keys())
        )
        umap_min_dist = st.sidebar.number_input(
            "UMAP Minimum Distance", min_value=0.0, max_value=1.0, value=0.1
        )

        show_vocab_overlap = st.sidebar.checkbox("Calculate Vocabulary Overlap")

        cluster_when = "before"
        clusterer = None
        if st.sidebar.checkbox("Cluster Embeddings"):
            clustering_algorithm = st.sidebar.selectbox(
                "Clustering Algorithm", ["K-Means", "HDBSCAN"]
            )
            cluster_when_labels = {
                "before": "Before Dimensionality Reduction",
                "after": "After Dimensionality Reduction",
            }
            cluster_when = st.sidebar.selectbox(
                "When to Apply Clusters",
                ["before", "after"],
                format_func=lambda l: cluster_when_labels[l],
            )
            if clustering_algorithm == "K-Means":
                n_clusters = st.sidebar.number_input(
                    "K-Means Number of Clusters", min_value=1, max_value=None, value=10
                )
                max_iter = st.sidebar.number_input(
                    "K-Means Max Iterations", min_value=1, max_value=None, value=300
                )

                kmeans_seed = st.sidebar.number_input(
                    "K-Means Random Seed", min_value=1, max_value=None, value=1
                )
                clusterer = KMeans(
                    n_clusters=n_clusters, max_iter=max_iter, random_state=kmeans_seed
                )
            else:
                hdbscan_min_cluster_size = st.sidebar.number_input(
                    "HDBSCAN Minimum Cluster Size", min_value=1, max_value=None, value=2
                )
                hdbscan_min_samples = st.sidebar.number_input(
                    "HDBSCAN Minimum Samples", min_value=1, max_value=None, value=2
                )
                hdbscan_cluster_selection_epsilon = st.sidebar.number_input(
                    "HDBSCAN Cluster Selection Epsilon",
                    min_value=0.0,
                    max_value=1.0,
                    value=0.1,
                )
                # There's a list of these available directly in HDBSCAN, but it's not
                # necessarily up-to-date with the distance metrics that are actually
                # implemented in sklearn
                hdbscan_metric_options = SKLEARN_DISTANCE_METRICS["ball_tree"]
                hdbscan_metric = st.sidebar.selectbox(
                    "HDBSCAN Distance Metric",
                    hdbscan_metric_options,
                    index=hdbscan_metric_options.index("euclidean"),
                )

                clusterer = hdbscan.HDBSCAN(
                    min_cluster_size=hdbscan_min_cluster_size,
                    min_samples=hdbscan_min_samples,
                    cluster_selection_epsilon=hdbscan_cluster_selection_epsilon,
                    metric=hdbscan_metric,
                )

    #
    # Main section
    #

    show_example_documents(
        sampled_texts,
        sampled_labels,
        filter_label,
        example_truncate_len,
        example_num_docs,
    )

    if filter_label is None:
        show_label_distribution(
            sampled_labels,
            all_labels=filtered_labels if show_full_label_distribution else None,
        )

    tokens = get_tokens(sampled_texts, tokenize_method, vocab_size)

    show_document_length_distribution(tokens)

    st.header("Topic Model")
    if not run_topic_model:
        st.markdown(
            "Enable topic modeling in the sidebar and click the 'Train Topic Model' button to train a topic model."
        )
    else:
        try:
            show_topic_model(
                tokens,
                sampled_labels,
                filter_label,
                num_topics=num_topics,
                train_chunksize=train_chunksize,
                train_passes=train_passes,
                train_iterations=train_iterations,
                do_bigrams=do_bigrams,
                bigram_min_count=bigram_min_count,
                min_frequency=min_frequency,
                max_proportion=max_proportion,
            )
        except ImportError as e:
            st.error(str(e))

    st.header("Embeddings")

    if embeddings_error:
        st.error(
            "Embeddings could not be generated due to bad parameters.  Look for errors in the 'Embeddings' section of the sidebar."
        )
    elif not run_embeddings:
        st.markdown(
            "Enable embeddings in the sidebar and click the 'Generate Embeddings' button to generate embeddings for the dataset."
        )
    else:
        show_embeddings(
            model_cls,
            model_kwargs,
            sampled_texts,
            sampled_labels,
            checkpoint_meta,
            batch_size=embed_batch_size,
            umap_seed=umap_seed,
            umap_n_neighbors=umap_n_neighbors,
            umap_metric=umap_metric,
            umap_min_dist=umap_min_dist,
            clusterer=clusterer,
            cluster_when=cluster_when,
            show_vocab_overlap=show_vocab_overlap,
        )


if __name__ == "__main__":
    # Streamlit doesn't like click exiting the script early after the function runs
    try:
        run()
    except SystemExit:
        pass



# Copyright (c) Microsoft. All rights reserved.
import math
import torch
from torch.optim import Optimizer
from torch.nn.utils import clip_grad_norm_
from pytorch_pretrained_bert.optimization import warmup_constant, warmup_cosine, warmup_linear

def warmup_linear_xdl(x, warmup=0.002):
    if x < warmup:
        return x/warmup
    return (1.0 - x)/(1.0 - warmup)

def schedule_func(sch):
    try:
        f = eval(sch)
    except:
        f = warmup_linear
    return f

class Adamax(Optimizer):
    """Implements BERT version of Adam algorithm with weight decay fix (and no ).
    Params:
        lr: learning rate
        warmup: portion of t_total for the warmup, -1  means no warmup. Default: -1
        t_total: total number of training steps for the learning
            rate schedule, -1  means constant learning rate. Default: -1
        schedule: schedule to use for the warmup (see above). Default: 'warmup_linear'
        b1: Adams b1. Default: 0.9
        b2: Adams b2. Default: 0.999
        e: Adams epsilon. Default: 1e-6
        weight_decay_rate: Weight decay. Default: 0.01
        max_grad_norm: Maximum norm for the gradients (-1 means no clipping). Default: 1.0
    by xiaodl 
    """
    def __init__(self, params, lr, warmup=-1, t_total=-1, schedule='warmup_linear',
                 betas=(0.9, 0.999), eps=1e-6, weight_decay_rate=0.01,
                 max_grad_norm=1.0):
        if not lr >= 0.0:
            raise ValueError("Invalid learning rate: {} - should be >= 0.0".format(lr))
        if not 0.0 <= warmup < 1.0 and not warmup == -1:
            raise ValueError("Invalid warmup: {} - should be in [0.0, 1.0[ or -1".format(warmup))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(lr=lr, schedule=schedule, warmup=warmup, t_total=t_total,
                        betas=betas, eps=eps, weight_decay_rate=weight_decay_rate,
                        max_grad_norm=max_grad_norm)
        super(Adamax, self).__init__(params, defaults)

    def get_lr(self):
        lr = []
        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                if len(state) == 0:
                    return [0]
                if group['t_total'] != -1:
                    schedule_fct = schedule_func(group['schedule'])
                    lr_scheduled = group['lr'] * schedule_fct(state['step']/group['t_total'], group['warmup'])
                else:
                    lr_scheduled = group['lr']
                lr.append(lr_scheduled)
        return lr

    def to(self, device):
        """ Move the optimizer state to a specified device"""
        for state in self.state.values():
            state['exp_avg'].to(device)
            state['exp_avg_sq'].to(device)

    def initialize_step(self, initial_step):
        """Initialize state with a defined step (but we don't have stored averaged).
        Arguments:
            initial_step (int): Initial step number.
        """
        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                # State initialization
                state['step'] = initial_step
                # Exponential moving average of gradient values
                state['exp_avg'] = torch.zeros_like(p.data)
                # Exponential moving average of squared gradient values
                state['exp_avg_sq'] = torch.zeros_like(p.data)

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_inf'] = torch.zeros_like(p.data)

                exp_avg, exp_inf = state['exp_avg'], state['exp_inf']
                beta1, beta2 = group['betas']
                eps = group['eps']
                # Add grad clipping
                if group['max_grad_norm'] > 0:
                    clip_grad_norm_(p, group['max_grad_norm'])

                # Update biased first moment estimate.
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                # Update the exponentially weighted infinity norm.
                norm_buf = torch.cat([
                    exp_inf.mul_(beta2).unsqueeze(0),
                    grad.abs().add_(eps).unsqueeze_(0)
                ], 0)
                torch.max(norm_buf, 0, keepdim=False, out=(exp_inf, exp_inf.new().long()))
                update = exp_avg / (exp_inf + eps)

                if group['weight_decay_rate'] > 0.0:
                    update += group['weight_decay_rate'] * p.data

                if group['t_total'] != -1:
                    schedule_fct = schedule_func(group['schedule'])
                    lr_scheduled = group['lr'] * schedule_fct(state['step']/group['t_total'], group['warmup'])
                else:
                    lr_scheduled = group['lr']

                update_with_lr = lr_scheduled * update
                p.data.add_(-update_with_lr)
                state['step'] += 1

        return loss



# coding=utf-8
# Copyright (c) Microsoft. All rights reserved.
import os
import torch
import math
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from module.dropout_wrapper import DropoutWrapper
from pytorch_pretrained_bert.modeling import BertConfig, BertEncoder, BertLayerNorm, BertModel
from module.san import SANClassifier, Classifier

class SANBertNetwork(nn.Module):
    def __init__(self, opt, bert_config=None):
        super(SANBertNetwork, self).__init__()
        self.dropout_list = nn.ModuleList()
        self.bert_config = BertConfig.from_dict(opt)
        self.bert = BertModel(self.bert_config)
        if opt['update_bert_opt'] > 0:
            for p in self.bert.parameters():
                p.requires_grad = False
        mem_size = self.bert_config.hidden_size
        self.decoder_opt = opt['answer_opt']
        self.scoring_list = nn.ModuleList()
        labels = [int(ls) for ls in opt['label_size'].split(',')]
        task_dropout_p = opt['tasks_dropout_p']
        self.bert_pooler = None

        for task, lab in enumerate(labels):
            decoder_opt = self.decoder_opt[task]
            dropout = DropoutWrapper(task_dropout_p[task], opt['vb_dropout'])
            self.dropout_list.append(dropout)
            if decoder_opt == 1:
                out_proj = SANClassifier(mem_size, mem_size, lab, opt, prefix='answer', dropout=dropout)
                self.scoring_list.append(out_proj)
            else:
                out_proj = nn.Linear(self.bert_config.hidden_size, lab)
                self.scoring_list.append(out_proj)

        self.opt = opt
        self._my_init()
        self.set_embed(opt)

    def _my_init(self):
        def init_weights(module):
            if isinstance(module, (nn.Linear, nn.Embedding)):
                # Slightly different from the TF version which uses truncated_normal for initialization
                # cf https://github.com/pytorch/pytorch/pull/5617
                module.weight.data.normal_(mean=0.0, std=self.bert_config.initializer_range * self.opt['init_ratio'])
            elif isinstance(module, BertLayerNorm):
                # Slightly different from the BERT pytorch version, which should be a bug.
                # Note that it only affects on training from scratch. For detailed discussions, please contact xiaodl@.
                # Layer normalization (https://arxiv.org/abs/1607.06450)
                # support both old/latest version
                if 'beta' in dir(module) and 'gamma' in dir(module):
                    module.beta.data.zero_()
                    module.gamma.data.fill_(1.0)
                else:
                    module.bias.data.zero_()
                    module.weight.data.fill_(1.0)
            if isinstance(module, nn.Linear):
                module.bias.data.zero_()
        self.apply(init_weights)

    def nbert_layer(self):
        return len(self.bert.encoder.layer)

    def freeze_layers(self, max_n):
        assert max_n < self.nbert_layer()
        for i in range(0, max_n):
            self.freeze_layer(i)

    def freeze_layer(self, n):
        assert n < self.nbert_layer()
        layer = self.bert.encoder.layer[n]
        for p in layer.parameters():
            p.requires_grad = False

    def set_embed(self, opt):
        bert_embeddings = self.bert.embeddings
        emb_opt = opt['embedding_opt']
        if emb_opt == 1:
            for p in bert_embeddings.word_embeddings.parameters():
                p.requires_grad = False
        elif emb_opt == 2:
            for p in bert_embeddings.position_embeddings.parameters():
                p.requires_grad = False
        elif emb_opt == 3:
            for p in bert_embeddings.token_type_embeddings.parameters():
                p.requires_grad = False
        elif emb_opt == 4:
            for p in bert_embeddings.token_type_embeddings.parameters():
                p.requires_grad = False
            for p in bert_embeddings.position_embeddings.parameters():
                p.requires_grad = False

    def forward(self, input_ids, token_type_ids, attention_mask, premise_mask=None, hyp_mask=None, task_id=0):
        all_encoder_layers, pooled_output = self.bert(input_ids, token_type_ids, attention_mask)
        sequence_output = all_encoder_layers[-1]
        if self.bert_pooler is not None:
            pooled_output = self.bert_pooler(sequence_output)
        decoder_opt = self.decoder_opt[task_id]
        if decoder_opt == 1:
            max_query = hyp_mask.size(1)
            assert max_query > 0
            assert premise_mask is not None
            assert hyp_mask is not None
            hyp_mem = sequence_output[:,:max_query,:]
            logits = self.scoring_list[task_id](sequence_output, hyp_mem, premise_mask, hyp_mask)
        else:
            pooled_output = self.dropout_list[task_id](pooled_output)
            logits = self.scoring_list[task_id](pooled_output)
        return logits



# Copyright (c) Microsoft. All rights reserved.
import os
import json
import tqdm
import pickle
import re
import collections
import argparse
from sys import path
from data_utils.vocab import Vocabulary
from pytorch_pretrained_bert.tokenization import BertTokenizer
from data_utils.log_wrapper import create_logger
from data_utils.label_map import GLOBAL_MAP
from data_utils.glue_utils import *
DEBUG_MODE=False
MAX_SEQ_LEN = 512

bert_tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
logger = create_logger(__name__, to_disk=True, log_file='bert_data_proc_512.log')

def _truncate_seq_pair(tokens_a, tokens_b, max_length):
    """Truncates a sequence pair in place to the maximum length.
    Copyed from https://github.com/huggingface/pytorch-pretrained-BERT
    """
    # This is a simple heuristic which will always truncate the longer sequence
    # one token at a time. This makes more sense than truncating an equal percent
    # of tokens from each, since if one sequence is very short then each token
    # that's truncated likely contains more information than a longer sequence.
    while True:
        total_length = len(tokens_a) + len(tokens_b)
        if total_length <= max_length:
            break
        if len(tokens_a) > len(tokens_b):
            tokens_a.pop()
        else:
            tokens_b.pop()

def build_data(data, dump_path, max_seq_len=MAX_SEQ_LEN, is_train=True, tolower=True):
    """Build data of sentence pair tasks
    """
    with open(dump_path, 'w', encoding='utf-8') as writer:
        for idx, sample in enumerate(data):
            ids = sample['uid']
            premise = bert_tokenizer.tokenize(sample['premise'])
            hypothesis = bert_tokenizer.tokenize(sample['hypothesis'])
            label = sample['label']
            _truncate_seq_pair(premise, hypothesis, max_seq_len - 3)
            input_ids =bert_tokenizer.convert_tokens_to_ids(['[CLS]'] + hypothesis + ['[SEP]'] + premise + ['[SEP]'])
            type_ids = [0] * ( len(hypothesis) + 2) + [1] * (len(premise) + 1)
            features = {'uid': ids, 'label': label, 'token_id': input_ids, 'type_id': type_ids}
            writer.write('{}\n'.format(json.dumps(features)))

def build_qnli(data, dump_path, max_seq_len=MAX_SEQ_LEN, is_train=True, tolower=True, gold_path=None):
    """Build QNLI as a pair-wise ranking task
    """
    with open(dump_path, 'w', encoding='utf-8') as writer:
        for idx, sample in enumerate(data):
            ids = sample['uid']
            premise = bert_tokenizer.tokenize(sample['premise'])
            hypothesis_1 = bert_tokenizer.tokenize(sample['hypothesis'][0])
            hypothesis_2 = bert_tokenizer.tokenize(sample['hypothesis'][1])
            label = sample['label']
            _truncate_seq_pair(premise, hypothesis_1, max_seq_len - 3)
            input_ids_1 =bert_tokenizer.convert_tokens_to_ids(['[CLS]'] + hypothesis_1 + ['[SEP]'] + premise + ['[SEP]'])
            type_ids_1 = [0] * ( len(hypothesis_1) + 2) + [1] * (len(premise) + 1)
            _truncate_seq_pair(premise, hypothesis_2, max_seq_len - 3)
            input_ids_2 =bert_tokenizer.convert_tokens_to_ids(['[CLS]'] + hypothesis_2 + ['[SEP]'] + premise + ['[SEP]'])
            type_ids_2 = [0] * ( len(hypothesis_2) + 2) + [1] * (len(premise) + 1)
            features = {'uid': ids, 'label': label, 'token_id': [input_ids_1, input_ids_2], 'type_id': [type_ids_1, type_ids_2], 'ruid':sample['ruid'], 'olabel':sample['olabel']}
            writer.write('{}\n'.format(json.dumps(features)))

def build_data_single(data, dump_path, max_seq_len=MAX_SEQ_LEN):
    """Build data of single sentence tasks
    """
    with open(dump_path, 'w', encoding='utf-8') as writer:
        for idx, sample in enumerate(data):
            ids = sample['uid']
            premise = bert_tokenizer.tokenize(sample['premise'])
            label = sample['label']
            if len(premise) >  max_seq_len - 3:
                premise = premise[:max_seq_len - 3] 
            input_ids =bert_tokenizer.convert_tokens_to_ids(['[CLS]'] + premise + ['[SEP]'])
            type_ids = [0] * ( len(premise) + 2)
            features = {'uid': ids, 'label': label, 'token_id': input_ids, 'type_id': type_ids}
            writer.write('{}\n'.format(json.dumps(features)))

def parse_args():
    parser = argparse.ArgumentParser(description='Preprocessing GLUE/SNLI/SciTail dataset.')
    parser.add_argument('--seed', type=int, default=13)
    parser.add_argument('--root_dir', type=str, default='data')
    parser.add_argument('--old_glue', action='store_true', help='whether it is old GLUE, refer official GLUE webpage for details')
    args = parser.parse_args()
    return args

def main(args):
    root = args.root_dir
    assert os.path.exists(root)
    is_old_glue = args.old_glue
    ######################################
    # SNLI/SciTail Tasks
    ######################################
    scitail_train_path = os.path.join(root, 'SciTail/tsv_format/scitail_1.0_train.tsv')
    scitail_dev_path = os.path.join(root, 'SciTail/tsv_format/scitail_1.0_dev.tsv')
    scitail_test_path = os.path.join(root, 'SciTail/tsv_format/scitail_1.0_test.tsv')

    snli_train_path = os.path.join(root, 'SNLI/train.tsv')
    snli_dev_path = os.path.join(root, 'SNLI/dev.tsv')
    snli_test_path = os.path.join(root, 'SNLI/test.tsv')

    ######################################
    # GLUE tasks
    ######################################
    multi_train_path =  os.path.join(root, 'MNLI/train.tsv')
    multi_dev_matched_path = os.path.join(root, 'MNLI/dev_matched.tsv')
    multi_dev_mismatched_path = os.path.join(root, 'MNLI/dev_mismatched.tsv')
    multi_test_matched_path = os.path.join(root, 'MNLI/test_matched.tsv')
    multi_test_mismatched_path = os.path.join(root, 'MNLI/test_mismatched.tsv')

    mrpc_train_path = os.path.join(root, 'MRPC/train.tsv')
    mrpc_dev_path = os.path.join(root, 'MRPC/dev.tsv')
    mrpc_test_path = os.path.join(root, 'MRPC/test.tsv')

    qnli_train_path = os.path.join(root, 'QNLI/train.tsv')
    qnli_dev_path = os.path.join(root, 'QNLI/dev.tsv')
    qnli_test_path = os.path.join(root, 'QNLI/test.tsv')

    qqp_train_path = os.path.join(root, 'QQP/train.tsv')
    qqp_dev_path = os.path.join(root, 'QQP/dev.tsv')
    qqp_test_path = os.path.join(root, 'QQP/test.tsv')

    rte_train_path = os.path.join(root, 'RTE/train.tsv')
    rte_dev_path = os.path.join(root, 'RTE/dev.tsv')
    rte_test_path = os.path.join(root, 'RTE/test.tsv')

    wnli_train_path = os.path.join(root, 'WNLI/train.tsv')
    wnli_dev_path = os.path.join(root, 'WNLI/dev.tsv')
    wnli_test_path = os.path.join(root, 'WNLI/test.tsv')

    stsb_train_path = os.path.join(root, 'STS-B/train.tsv')
    stsb_dev_path = os.path.join(root, 'STS-B/dev.tsv')
    stsb_test_path = os.path.join(root, 'STS-B/test.tsv')

    sst_train_path = os.path.join(root, 'SST-2/train.tsv')
    sst_dev_path = os.path.join(root, 'SST-2/dev.tsv')
    sst_test_path = os.path.join(root, 'SST-2/test.tsv')

    cola_train_path = os.path.join(root, 'CoLA/train.tsv')
    cola_dev_path = os.path.join(root, 'CoLA/dev.tsv')
    cola_test_path = os.path.join(root, 'CoLA/test.tsv')
    
    ######################################
    # Loading DATA
    ######################################
    scitail_train_data = load_scitail(scitail_train_path, GLOBAL_MAP['scitail'])
    scitail_dev_data = load_scitail(scitail_dev_path, GLOBAL_MAP['scitail'])
    scitail_test_data = load_scitail(scitail_test_path, GLOBAL_MAP['scitail'])
    logger.info('Loaded {} SciTail train samples'.format(len(scitail_train_data)))
    logger.info('Loaded {} SciTail dev samples'.format(len(scitail_dev_data)))
    logger.info('Loaded {} SciTail test samples'.format(len(scitail_test_data)))

    snli_train_data = load_snli(snli_train_path, GLOBAL_MAP['snli'])
    snli_dev_data = load_snli(snli_dev_path, GLOBAL_MAP['snli'])
    snli_test_data = load_snli(snli_test_path, GLOBAL_MAP['snli'])
    logger.info('Loaded {} SNLI train samples'.format(len(snli_train_data)))
    logger.info('Loaded {} SNLI dev samples'.format(len(snli_dev_data)))
    logger.info('Loaded {} SNLI test samples'.format(len(snli_test_data)))

    multinli_train_data = load_mnli(multi_train_path, GLOBAL_MAP['snli'])
    multinli_matched_dev_data = load_mnli(multi_dev_matched_path, GLOBAL_MAP['snli'])
    multinli_mismatched_dev_data = load_mnli(multi_dev_mismatched_path, GLOBAL_MAP['snli'])
    multinli_matched_test_data = load_mnli(multi_test_matched_path, GLOBAL_MAP['snli'], is_train=False)
    multinli_mismatched_test_data = load_mnli(multi_test_mismatched_path, GLOBAL_MAP['snli'], is_train=False)

    logger.info('Loaded {} MNLI train samples'.format(len(multinli_train_data)))
    logger.info('Loaded {} MNLI matched dev samples'.format(len(multinli_matched_dev_data)))
    logger.info('Loaded {} MNLI mismatched dev samples'.format(len(multinli_mismatched_dev_data)))
    logger.info('Loaded {} MNLI matched test samples'.format(len(multinli_matched_test_data)))
    logger.info('Loaded {} MNLI mismatched test samples'.format(len(multinli_mismatched_test_data)))

    mrpc_train_data = load_mrpc(mrpc_train_path)
    mrpc_dev_data = load_mrpc(mrpc_dev_path)
    mrpc_test_data = load_mrpc(mrpc_test_path, is_train=False)
    logger.info('Loaded {} MRPC train samples'.format(len(mrpc_train_data)))
    logger.info('Loaded {} MRPC dev samples'.format(len(mrpc_dev_data)))
    logger.info('Loaded {} MRPC test samples'.format(len(mrpc_test_data)))

    qnli_train_data = load_qnli(qnli_train_path, GLOBAL_MAP['qnli'])
    qnli_dev_data = load_qnli(qnli_dev_path, GLOBAL_MAP['qnli'])
    qnli_test_data = load_qnli(qnli_test_path, GLOBAL_MAP['qnli'], is_train=False)
    logger.info('Loaded {} QNLI train samples'.format(len(qnli_train_data)))
    logger.info('Loaded {} QNLI dev samples'.format(len(qnli_dev_data)))
    logger.info('Loaded {} QNLI test samples'.format(len(qnli_test_data)))
    
    if is_old_glue:
        qnnli_train_data = load_qnnli(qnli_train_path, GLOBAL_MAP['qnli'])
        qnnli_dev_data = load_qnnli(qnli_dev_path, GLOBAL_MAP['qnli'])
        qnnli_test_data = load_qnnli(qnli_test_path, GLOBAL_MAP['qnli'], is_train=False)
        logger.info('Loaded {} QNLI train samples'.format(len(qnli_train_data)))
        logger.info('Loaded {} QNLI dev samples'.format(len(qnli_dev_data)))
        logger.info('Loaded {} QNLI test samples'.format(len(qnli_test_data)))

    qqp_train_data = load_qqp(qqp_train_path)
    qqp_dev_data = load_qqp(qqp_dev_path)
    qqp_test_data = load_qqp(qqp_test_path, is_train=False)
    logger.info('Loaded {} QQP train samples'.format(len(qqp_train_data)))
    logger.info('Loaded {} QQP dev samples'.format(len(qqp_dev_data)))
    logger.info('Loaded {} QQP test samples'.format(len(qqp_test_data)))

    rte_train_data = load_rte(rte_train_path, GLOBAL_MAP['rte'])
    rte_dev_data = load_rte(rte_dev_path, GLOBAL_MAP['rte'])
    rte_test_data = load_rte(rte_test_path, GLOBAL_MAP['rte'], is_train=False)
    logger.info('Loaded {} RTE train samples'.format(len(rte_train_data)))
    logger.info('Loaded {} RTE dev samples'.format(len(rte_dev_data)))
    logger.info('Loaded {} RTE test samples'.format(len(rte_test_data)))

    wnli_train_data = load_wnli(wnli_train_path)
    wnli_dev_data = load_wnli(wnli_dev_path)
    wnli_test_data = load_wnli(wnli_test_path, is_train=False)
    logger.info('Loaded {} WNLI train samples'.format(len(wnli_train_data)))
    logger.info('Loaded {} WNLI dev samples'.format(len(wnli_dev_data)))
    logger.info('Loaded {} WNLI test samples'.format(len(wnli_test_data)))

    sst_train_data = load_sst(sst_train_path)
    sst_dev_data = load_sst(sst_dev_path)
    sst_test_data = load_sst(sst_test_path, is_train=False)
    logger.info('Loaded {} SST train samples'.format(len(sst_train_data)))
    logger.info('Loaded {} SST dev samples'.format(len(sst_dev_data)))
    logger.info('Loaded {} SST test samples'.format(len(sst_test_data)))

    stsb_train_data = load_sts(stsb_train_path)
    stsb_dev_data = load_sts(stsb_dev_path)
    stsb_test_data = load_sts(stsb_test_path, is_train=False)
    logger.info('Loaded {} STS-B train samples'.format(len(stsb_train_data)))
    logger.info('Loaded {} STS-B dev samples'.format(len(stsb_dev_data)))
    logger.info('Loaded {} STS-B test samples'.format(len(stsb_test_data)))

    cola_train_data = load_cola(cola_train_path)
    cola_dev_data = load_cola(cola_dev_path)
    cola_test_data = load_cola(cola_test_path, is_train=False)
    logger.info('Loaded {} COLA train samples'.format(len(cola_train_data)))
    logger.info('Loaded {} COLA dev samples'.format(len(cola_dev_data)))
    logger.info('Loaded {} COLA test samples'.format(len(cola_test_data)))

    mt_dnn_root = os.path.join(root, 'mt_dnn')
    if not os.path.isdir(mt_dnn_root):
        os.mkdir(mt_dnn_root)

    # BUILD SciTail
    scitail_train_fout = os.path.join(mt_dnn_root, 'scitail_train.json')
    scitail_dev_fout = os.path.join(mt_dnn_root, 'scitail_dev.json')
    scitail_test_fout = os.path.join(mt_dnn_root, 'scitail_test.json')
    build_data(scitail_train_data, scitail_train_fout)
    build_data(scitail_dev_data, scitail_dev_fout)
    build_data(scitail_test_data, scitail_test_fout)
    logger.info('done with scitail')

    # BUILD SNLI
    snli_train_fout = os.path.join(mt_dnn_root, 'snli_train.json')
    snli_dev_fout = os.path.join(mt_dnn_root, 'snli_dev.json')
    snli_test_fout = os.path.join(mt_dnn_root, 'snli_test.json')
    build_data(snli_train_data, snli_train_fout)
    build_data(snli_dev_data, snli_dev_fout)
    build_data(snli_test_data, snli_test_fout)
    logger.info('done with snli')

    # BUILD MNLI
    multinli_train_fout = os.path.join(mt_dnn_root, 'mnli_train.json')
    multinli_matched_dev_fout = os.path.join(mt_dnn_root, 'mnli_matched_dev.json')
    multinli_mismatched_dev_fout = os.path.join(mt_dnn_root, 'mnli_mismatched_dev.json')
    multinli_matched_test_fout = os.path.join(mt_dnn_root, 'mnli_matched_test.json')
    multinli_mismatched_test_fout = os.path.join(mt_dnn_root, 'mnli_mismatched_test.json')
    build_data(multinli_train_data, multinli_train_fout)
    build_data(multinli_matched_dev_data, multinli_matched_dev_fout)
    build_data(multinli_mismatched_dev_data, multinli_mismatched_dev_fout)
    build_data(multinli_matched_test_data, multinli_matched_test_fout)
    build_data(multinli_mismatched_test_data, multinli_mismatched_test_fout)
    logger.info('done with mnli')

    mrpc_train_fout = os.path.join(mt_dnn_root, 'mrpc_train.json')
    mrpc_dev_fout = os.path.join(mt_dnn_root, 'mrpc_dev.json')
    mrpc_test_fout = os.path.join(mt_dnn_root, 'mrpc_test.json')
    build_data(mrpc_train_data, mrpc_train_fout)
    build_data(mrpc_dev_data, mrpc_dev_fout)
    build_data(mrpc_test_data, mrpc_test_fout)
    logger.info('done with mrpc')

    qnli_train_fout = os.path.join(mt_dnn_root, 'qnli_train.json')
    qnli_dev_fout = os.path.join(mt_dnn_root, 'qnli_dev.json')
    qnli_test_fout = os.path.join(mt_dnn_root, 'qnli_test.json')
    build_data(qnli_train_data, qnli_train_fout)
    build_data(qnli_dev_data, qnli_dev_fout)
    build_data(qnli_test_data, qnli_test_fout)
    logger.info('done with qnli')

    if is_old_glue:
        qnli_train_fout = os.path.join(mt_dnn_root, 'qnnli_train.json')
        qnli_dev_fout = os.path.join(mt_dnn_root, 'qnnli_dev.json')
        qnli_test_fout = os.path.join(mt_dnn_root, 'qnnli_test.json')
        qnli_dev_gold_fout = os.path.join(mt_dnn_root, 'qnli_gold_dev.tsv')
        build_qnli(qnnli_train_data, qnli_train_fout)
        build_qnli(qnnli_dev_data, qnli_dev_fout)
        build_qnli(qnnli_train_data, qnli_test_fout)
        logger.info('done with qnli')

    qqp_train_fout = os.path.join(mt_dnn_root, 'qqp_train.json')
    qqp_dev_fout = os.path.join(mt_dnn_root, 'qqp_dev.json')
    qqp_test_fout = os.path.join(mt_dnn_root, 'qqp_test.json')
    build_data(qqp_train_data, qqp_train_fout)
    build_data(qqp_dev_data, qqp_dev_fout)
    build_data(qqp_test_data, qqp_test_fout)
    logger.info('done with qqp')

    rte_train_fout = os.path.join(mt_dnn_root, 'rte_train.json')
    rte_dev_fout = os.path.join(mt_dnn_root, 'rte_dev.json')
    rte_test_fout = os.path.join(mt_dnn_root, 'rte_test.json')
    build_data(rte_train_data, rte_train_fout)
    build_data(rte_dev_data, rte_dev_fout)
    build_data(rte_test_data, rte_test_fout)
    logger.info('done with rte')

    wnli_train_fout = os.path.join(mt_dnn_root, 'wnli_train.json')
    wnli_dev_fout = os.path.join(mt_dnn_root, 'wnli_dev.json')
    wnli_test_fout = os.path.join(mt_dnn_root, 'wnli_test.json')
    build_data(wnli_train_data, wnli_train_fout)
    build_data(wnli_dev_data, wnli_dev_fout)
    build_data(wnli_test_data, wnli_test_fout)
    logger.info('done with wnli')

    stsb_train_fout = os.path.join(mt_dnn_root, 'stsb_train.json')
    stsb_dev_fout = os.path.join(mt_dnn_root, 'stsb_dev.json')
    stsb_test_fout = os.path.join(mt_dnn_root, 'stsb_test.json')
    build_data(stsb_train_data, stsb_train_fout)
    build_data(stsb_dev_data, stsb_dev_fout)
    build_data(stsb_test_data, stsb_test_fout)
    logger.info('done with stsb')

    sst_train_fout = os.path.join(mt_dnn_root, 'sst_train.json')
    sst_dev_fout = os.path.join(mt_dnn_root, 'sst_dev.json')
    sst_test_fout = os.path.join(mt_dnn_root, 'sst_test.json')
    build_data_single(sst_train_data, sst_train_fout)
    build_data_single(sst_dev_data, sst_dev_fout)
    build_data_single(sst_test_data, sst_test_fout)
    logger.info('done with sst')

    cola_train_fout = os.path.join(mt_dnn_root, 'cola_train.json')
    cola_dev_fout = os.path.join(mt_dnn_root, 'cola_dev.json')
    cola_test_fout = os.path.join(mt_dnn_root, 'cola_test.json')
    build_data_single(cola_train_data, cola_train_fout)
    build_data_single(cola_dev_data, cola_dev_fout)
    build_data_single(cola_test_data, cola_test_fout)
    logger.info('done with cola')

if __name__ == '__main__':
    args = parse_args()
    main(args)



import warnings
from pathlib import Path
from typing import Any, Dict, Optional

import joblib
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin, is_classifier
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score, f1_score
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline

import gobbli.io
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import EmbedMixin, PredictMixin, TrainMixin
from gobbli.util import assert_type, generate_uuid


def persist_estimator(estimator: BaseEstimator) -> Path:
    """
    Saves the given estimator to a gobbli-managed filepath, where it can be loaded from
    disk by the SKLearnClassifier.  This is useful if you want to use an estimator but
    don't want to bother with saving it to disk on your own.

    Args:
      estimator: The estimator to load.

    Returns:
      The path where the estimator was saved.
    """
    estimator_dir = (
        SKLearnClassifier.model_class_dir() / "user_estimators" / generate_uuid()
    )
    estimator_dir.mkdir(exist_ok=True, parents=True)

    estimator_path = estimator_dir / SKLearnClassifier._TRAIN_OUTPUT_CHECKPOINT
    SKLearnClassifier._dump_estimator(estimator, estimator_path)

    return estimator_path


def make_default_tfidf_logistic_regression() -> BaseEstimator:
    """
    Returns:
      A pipeline composing a TF-IDF vectorizer and a logistic regression model using
      default parameters.
    """
    return Pipeline(
        [("tfidf", TfidfVectorizer()), ("logreg", LogisticRegression(random_state=1))]
    )


def make_cv_tfidf_logistic_regression(
    grid_params: Optional[Dict[str, Any]] = None
) -> BaseEstimator:
    """
    Args:
      grid_params: Grid search parameters for the pipeline.  Passed directly to
        :class:`sklearn.model_selection.GridSearchCV`.  See
        :func:`make_default_tfidf_logistic_regression` for the names of the pipeline
        components.  If not given, will use a somewhat reasonable default.

    Returns:
      A cross-validated pipeline combining a TF-IDF vectorizer and logistic regression model
      with the specified grid parameters.
    """
    if grid_params is None:
        grid_params = {
            "tfidf__ngram_range": [(1, 2)],
            "tfidf__min_df": [0.01],
            "tfidf__max_df": [0.95],
            "logreg__C": [0.1, 0.5, 1],
            "logreg__l1_ratio": [0, 0.25, 0.5, 0.75, 1],
            "logreg__solver": ["saga"],
            "logreg__penalty": ["elasticnet"],
        }
    return GridSearchCV(
        make_default_tfidf_logistic_regression(),
        grid_params,
        cv=5,
        return_train_score=True,
        verbose=10,
    )


_AT_LEAST_TWO_CLASSES_ERR_MSG = (
    "This solver needs samples of at least 2 classes in the data"
)


class _SafeEstimator(BaseEstimator, ClassifierMixin):
    """
    Wrap an arbitrary classifier estimator to catch errors when fitting
    models that require more than 1 class in the data.
    """

    def __init__(self, base_estimator: BaseEstimator):
        self.base_estimator = base_estimator
        self.classes_: Optional[np.ndarray] = None

        if hasattr(base_estimator, "classes_"):
            self.classes_ = self.base_estimator.classes_

    def fit(self, *args, **kwargs):
        try:
            return self.base_estimator.fit(*args, **kwargs)
        except ValueError as e:
            if _AT_LEAST_TWO_CLASSES_ERR_MSG not in str(e):
                raise
        finally:
            self.classes_ = self.base_estimator.classes_

    def predict_proba(self, X):
        if self.classes_ is None:
            raise ValueError(
                "Can't predict without knowing what the estimator's classes are."
            )

        if len(self.classes_) == 1:
            return np.ones((len(X), 1))
        return self.base_estimator.predict_proba(X)

    def predict(self, X):
        if self.classes_ is None:
            raise ValueError(
                "Can't predict without knowing what the estimator's classes are."
            )

        if len(self.classes_) == 1:
            return np.full_like(X, self.classes_[0])
        return self.base_estimator.predict(X)


class SKLearnClassifier(BaseModel, TrainMixin, PredictMixin):
    """
    Classifier wrapper for `scikit-learn <https://scikit-learn.org/stable/>`__ classifiers.
    Wraps a :class:`sklearn.base.BaseEstimator` which accepts text input and outputs
    predictions.

    Creating an estimator that meets those conditions will generally require
    some use of :class:`sklearn.pipeline.Pipeline` to compose a transform (e.g. a vectorizer
    to vectorize text) and an estimator (e.g. logistic regression).  See the helper functions
    in this module for some examples.  You may also consider wrapping the pipeline with
    :class:`sklearn.model_selection.GridSearchCV` to tune hyperparameters.
    """

    _TRAIN_OUTPUT_CHECKPOINT = "estimator.joblib"

    def init(self, params: Dict[str, Any]):
        """
        See :meth:`gobbli.model.base.BaseModel.init`.

        SKLearnClassifier parameters:

        - ``estimator_path`` (:obj:`str`): Path to an estimator pickled by joblib.
          The pickle will be loaded, and the resulting object will be used as the estimator.
          If not provided, a default pipeline composed of a TF-IDF vectorizer and a
          logistic regression will be used.
        """
        estimator = None

        for name, value in params.items():
            if name == "estimator_path":
                assert_type(name, value, str)
                estimator = SKLearnClassifier._load_estimator(Path(value))
                SKLearnClassifier._validate_estimator(estimator)
            else:
                raise ValueError(f"Unknown param '{name}'")

        if estimator is None:
            self.estimator = _SafeEstimator(make_default_tfidf_logistic_regression())
        else:
            self.estimator = _SafeEstimator(estimator)

    @staticmethod
    def _load_estimator(estimator_path: Path) -> BaseEstimator:
        return joblib.load(estimator_path)

    @staticmethod
    def _dump_estimator(estimator: BaseEstimator, estimator_path: Path):
        joblib.dump(estimator, estimator_path)

    @staticmethod
    def _validate_estimator(estimator: BaseEstimator):
        """
        Run some checks on the given object to determine if it's an estimator which is
        valid for our purposes.
        """
        # sklearn has a function that does a lot more intensive checking regarding
        # the interface of a candidate Estimator
        # (sklearn.utils.estimator_checks.check_estimator), but the function
        # doesn't work well for our use case as of version 0.22.  It doesn't properly
        # detect Pipeline X_types based on the first pipeline component and won't
        # test anything that doesn't accept a 2-D numpy array as input.  We'll settle
        # for lax checks here until sklearn has something that works better for us.
        if not is_classifier(estimator):
            raise ValueError(
                "Estimator must be a classifier according to sklearn.base.is_classifier()"
            )

        if not hasattr(estimator, "predict_proba"):
            raise ValueError(
                "Estimator must support the predict_proba() method to fulfill gobbli's "
                "interface requirements for a prediction model."
            )

    def _build(self):
        """
        No build step required for this model.
        """

    def _train(
        self, train_input: gobbli.io.TrainInput, context: ContainerTaskContext
    ) -> gobbli.io.TrainOutput:
        if train_input.checkpoint is not None:
            warnings.warn(
                "SKLearnClassifier does not support training from an existing "
                "checkpoint, so the passed checkpoint will be ignored."
            )
        self.estimator.fit(train_input.X_train, train_input.y_train)

        y_train_pred = self.estimator.predict(train_input.X_train)
        train_loss = -f1_score(
            train_input.y_train, y_train_pred, zero_division="warn", average="weighted"
        )

        y_valid_pred = self.estimator.predict(train_input.X_valid)
        valid_loss = -f1_score(
            train_input.y_valid, y_valid_pred, zero_division="warn", average="weighted"
        )
        valid_accuracy = accuracy_score(train_input.y_valid, y_valid_pred)

        checkpoint_path = (
            context.host_output_dir / SKLearnClassifier._TRAIN_OUTPUT_CHECKPOINT
        )
        self._dump_estimator(self.estimator.base_estimator, checkpoint_path)

        return gobbli.io.TrainOutput(
            valid_loss=valid_loss,
            valid_accuracy=valid_accuracy,
            train_loss=train_loss,
            labels=train_input.labels(),
            checkpoint=checkpoint_path,
        )

    def _predict(
        self, predict_input: gobbli.io.PredictInput, context: ContainerTaskContext
    ) -> gobbli.io.PredictOutput:

        if predict_input.checkpoint is not None:
            self.estimator = _SafeEstimator(
                self._load_estimator(predict_input.checkpoint)
            )

        pred_proba_df = pd.DataFrame(self.estimator.predict_proba(predict_input.X))
        if self.estimator.classes_ is None:
            raise ValueError(
                "Can't determine column names for predicted probabilities."
            )
        pred_proba_df.columns = self.estimator.classes_.astype("str")

        labels = predict_input.labels
        for label in labels:
            if label not in pred_proba_df.columns:
                pred_proba_df[label] = 0.0

        return gobbli.io.PredictOutput(y_pred_proba=pred_proba_df)


class TfidfEmbedder(BaseModel, EmbedMixin):
    """
    Embedding wrapper for scikit-learn's :class:`sklearn.feature_extraction.text.TfidfVectorizer`.
    Generates "embeddings" composed of TF-IDF vectors.
    """

    def init(self, params: Dict[str, Any]):
        """
        See :meth:`gobbli.model.base.BaseModel.init`.

        TFidfEmbedder parameters will be passed directly to the
        :class:`sklearn.feature_extraction.text.TfidfVectorizer` constructor, which will
        perform its own validation.
        """
        # This should raise an error if there's anything wrong with any of the passed
        # parameters
        self.vec = TfidfVectorizer(**params)

    def _build(self):
        """
        No build step required for this model.
        """

    def _embed(
        self, embed_input: gobbli.io.EmbedInput, context: ContainerTaskContext
    ) -> gobbli.io.EmbedOutput:
        if embed_input.checkpoint is not None:
            warnings.warn(
                "TfidfEmbedder does not support embedding from an existing "
                "checkpoint, so the passed checkpoint will be ignored."
            )
        if embed_input.pooling == gobbli.io.EmbedPooling.NONE:
            raise ValueError(
                "TfidfEmbedder embeds whole documents, so pooling is required."
            )
        X_vectorized = self.vec.fit_transform(embed_input.X)

        return gobbli.io.EmbedOutput(
            X_embedded=[np.squeeze(np.asarray(vec.todense())) for vec in X_vectorized]
        )



import argparse
import csv
import json
import logging
import os
from datetime import datetime

import torch
from data_utils.log_wrapper import create_logger
from data_utils.metrics import compute_acc, compute_cross_entropy
from data_utils.utils import set_environment
from mt_dnn.gobbli_batcher import GobbliBatchGen
from mt_dnn.gobbli_model import GobbliMTDNNModel

logging.basicConfig(level=logging.INFO)


def model_config(parser):
    parser.add_argument("--update_bert_opt", default=0, type=int)
    parser.add_argument("--multi_gpu_on", action="store_true")
    parser.add_argument(
        "--mem_cum_type", type=str, default="simple", help="bilinear/simple/defualt"
    )
    parser.add_argument("--answer_num_turn", type=int, default=5)
    parser.add_argument("--answer_mem_drop_p", type=float, default=0.1)
    parser.add_argument("--answer_att_hidden_size", type=int, default=128)
    parser.add_argument(
        "--answer_att_type",
        type=str,
        default="bilinear",
        help="bilinear/simple/defualt",
    )
    parser.add_argument(
        "--answer_rnn_type", type=str, default="gru", help="rnn/gru/lstm"
    )
    parser.add_argument(
        "--answer_sum_att_type",
        type=str,
        default="bilinear",
        help="bilinear/simple/defualt",
    )
    parser.add_argument("--answer_merge_opt", type=int, default=1)
    parser.add_argument("--answer_mem_type", type=int, default=1)
    parser.add_argument("--answer_dropout_p", type=float, default=0.1)
    parser.add_argument("--answer_weight_norm_on", action="store_true")
    parser.add_argument("--dump_state_on", action="store_true")
    parser.add_argument("--answer_opt", type=int, default=0, help="0,1")
    parser.add_argument("--mtl_opt", type=int, default=0)
    parser.add_argument("--ratio", type=float, default=0)
    parser.add_argument("--mix_opt", type=int, default=0)
    parser.add_argument("--max_seq_len", type=int, default=512)
    parser.add_argument("--init_ratio", type=float, default=1)
    return parser


def data_config(parser):
    parser.add_argument(
        "--log_file", default="mt-dnn-train.log", help="path for log file."
    )
    parser.add_argument(
        "--init_checkpoint", default="mt_dnn/bert_model_base.pt", type=str
    )
    parser.add_argument("--data_dir", default="data/mt_dnn")
    parser.add_argument("--data_sort_on", action="store_true")
    parser.add_argument("--name", default="farmer")
    parser.add_argument("--train_file")
    parser.add_argument("--valid_file")
    parser.add_argument("--test_file")
    parser.add_argument("--label_file", required=True)
    return parser


def train_config(parser):
    parser.add_argument(
        "--cuda",
        type=bool,
        default=torch.cuda.is_available(),
        help="whether to use GPU acceleration.",
    )
    parser.add_argument("--log_per_updates", type=int, default=500)
    parser.add_argument("--epochs", type=int, default=5)
    parser.add_argument("--batch_size", type=int, default=8)
    parser.add_argument("--batch_size_eval", type=int, default=8)
    parser.add_argument(
        "--optimizer",
        default="adamax",
        help="supported optimizer: adamax, sgd, adadelta, adam",
    )
    parser.add_argument("--grad_clipping", type=float, default=0)
    parser.add_argument("--global_grad_clipping", type=float, default=1.0)
    parser.add_argument("--weight_decay", type=float, default=0)
    parser.add_argument("--learning_rate", type=float, default=5e-5)
    parser.add_argument("--momentum", type=float, default=0)
    parser.add_argument("--warmup", type=float, default=0.1)
    parser.add_argument("--warmup_schedule", type=str, default="warmup_linear")

    parser.add_argument("--vb_dropout", action="store_false")
    parser.add_argument("--dropout_p", type=float, default=0.1)
    parser.add_argument("--dropout_w", type=float, default=0.000)
    parser.add_argument("--bert_dropout_p", type=float, default=0.1)

    # EMA
    parser.add_argument("--ema_opt", type=int, default=0)
    parser.add_argument("--ema_gamma", type=float, default=0.995)

    # scheduler
    parser.add_argument(
        "--have_lr_scheduler", dest="have_lr_scheduler", action="store_false"
    )
    parser.add_argument("--multi_step_lr", type=str, default="10,20,30")
    parser.add_argument("--freeze_layers", type=int, default=-1)
    parser.add_argument("--embedding_opt", type=int, default=0)
    parser.add_argument("--lr_gamma", type=float, default=0.5)
    parser.add_argument("--bert_l2norm", type=float, default=0.0)
    parser.add_argument("--scheduler_type", type=str, default="ms", help="ms/rop/exp")
    parser.add_argument("--output_dir", default="checkpoint")
    parser.add_argument(
        "--seed",
        type=int,
        default=2018,
        help="random seed for data shuffling, embedding init, etc.",
    )
    parser.add_argument(
        "--task_config_path", type=str, default="configs/tasks_config.json"
    )

    return parser


def dump(path, data):
    with open(path, "w") as f:
        json.dump(data, f)


def dump_predictions(scores, labels, output_file):
    # Map label indices (used internally for predictions) to label names
    # (which the user will be familiar with)
    labels_inverse = {ndx: label for label, ndx in labels.items()}
    label_header = [labels_inverse[i] for i in range(len(labels))]
    with open(output_file, "w") as f:
        writer = csv.writer(f)
        # Write the header
        writer.writerow(label_header)

        # Write the predictions
        for row in scores.tolist():
            writer.writerow(row)


def read_labels(label_file):
    with open(label_file, "r") as f:
        return {l.strip(): i for i, l in enumerate(f.readlines())}


def eval_model(model, dataset, use_cuda=True, with_label=True):
    dataset.reset()
    if use_cuda:
        model.cuda()

    predictions = []
    scores = []
    golds = []
    metrics = {}
    for batch in dataset:
        if with_label:
            input_ids, token_type_ids, attention_mask, labels = batch
            golds.extend(labels)
        else:
            input_ids, token_type_ids, attention_mask = batch

        score, pred = model.predict(input_ids, token_type_ids, attention_mask)
        predictions.extend(pred)
        scores.append(score)

    scores = torch.cat(scores, 0)

    if not with_label:
        return predictions, scores

    metrics["accuracy"] = compute_acc(predictions, golds)
    metrics["loss"] = compute_cross_entropy(scores, torch.LongTensor(golds)).item()

    return metrics, predictions, scores


parser = argparse.ArgumentParser()
parser = data_config(parser)
parser = model_config(parser)
parser = train_config(parser)
args = parser.parse_args()

given_train = args.train_file is not None
given_valid = args.valid_file is not None
given_test = args.test_file is not None

if given_train and not given_valid or given_valid and not given_train:
    raise ValueError("Must have both a train and valid dataset for training.")

output_dir = args.output_dir
data_dir = args.data_dir

os.makedirs(output_dir, exist_ok=True)
output_dir = os.path.abspath(output_dir)

set_environment(args.seed, args.cuda)
log_path = args.log_file
logger = create_logger(__name__, to_disk=True, log_file=log_path)


def main():
    logger.info("Launching the MT-DNN training")
    opt = vars(args)
    # update data dir
    opt["data_dir"] = data_dir

    labels = read_labels(args.label_file)
    # The original code expects this to be a comma-separated list of ints in string format
    # to account for multiple tasks having different label sizes
    # It works out if we just use our single label size as a string
    opt["label_size"] = str(len(labels))
    # This option was also assigned per-task in the original code, but we'll just take the default
    opt["tasks_dropout_p"] = [args.dropout_p]
    # This was also per-task, and the default command line arg is a scalar instead of a list,
    # so set it correctly to be a list
    opt["answer_opt"] = [0]
    batch_size = args.batch_size

    batch_gen_kwargs = {
        "maxlen": args.max_seq_len,
        "batch_size": batch_size,
        "dropout_w": args.dropout_w,
        "gpu": args.cuda,
    }

    if given_train:
        train_data = GobbliBatchGen(
            args.train_file,
            has_labels=True,
            labels=labels,
            is_train=True,
            **batch_gen_kwargs,
        )
        num_all_batches = args.epochs * len(train_data)
    else:
        num_all_batches = 0

    if given_valid:
        valid_data = GobbliBatchGen(
            args.valid_file,
            has_labels=True,
            labels=labels,
            is_train=False,
            **batch_gen_kwargs,
        )

    if given_test:
        test_data = GobbliBatchGen(
            args.test_file, has_labels=False, is_train=False, **batch_gen_kwargs
        )

    model_path = args.init_checkpoint

    if os.path.exists(model_path):
        state_dict = torch.load(model_path)
        config = state_dict["config"]
        config["attention_probs_dropout_prob"] = args.bert_dropout_p
        config["hidden_dropout_prob"] = args.bert_dropout_p
        opt.update(config)
    else:
        raise ValueError("Could not find the init model!")

    model = GobbliMTDNNModel(opt, state_dict=state_dict, num_train_step=num_all_batches)

    ####model meta str
    headline = "############# Model Arch of MT-DNN #############"
    ###print network
    logger.info("\n{}\n{}\n".format(headline, model.network))

    # dump config
    config_file = os.path.join(output_dir, "config.json")
    with open(config_file, "w", encoding="utf-8") as writer:
        writer.write("{}\n".format(json.dumps(opt)))
        writer.write("\n{}\n{}\n".format(headline, model.network))

    logger.info("Total number of params: {}".format(model.total_param))

    if args.freeze_layers > 0:
        model.network.freeze_layers(args.freeze_layers)

    if args.cuda:
        model.cuda()

    if given_train and given_valid:
        for epoch in range(0, args.epochs):
            logger.warning("At epoch {}".format(epoch))
            start = datetime.now()

            # Training
            train_data.reset()
            for i, (input_ids, token_type_ids, attention_mask, labels) in enumerate(
                train_data
            ):
                model.update(input_ids, token_type_ids, attention_mask, labels)
                if (model.updates) % args.log_per_updates == 0 or model.updates == 1:
                    logger.info(
                        "updates[{0:6}] train loss[{1:.5f}] remaining[{2}]".format(
                            model.updates,
                            model.train_loss.avg,
                            str(
                                (datetime.now() - start)
                                / (i + 1)
                                * (len(train_data) - i - 1)
                            ).split(".")[0],
                        )
                    )

            # Training and validation metrics
            for dataset, name in ((train_data, "train"), (valid_data, "valid")):
                metrics, _, _ = eval_model(
                    model, dataset, use_cuda=args.cuda, with_label=True
                )

                for key, val in metrics.items():
                    logger.warning(
                        "Epoch {0} -- [{1}] {2}: {3:.3f}".format(epoch, name, key, val)
                    )

                score_file = os.path.join(
                    output_dir, "{}_scores_{}.json".format(name, epoch)
                )
                results = {"metrics": metrics}
                dump(score_file, results)
        model_file = os.path.join(output_dir, "model_{}.pt".format(epoch))
        model.save(model_file)

    if given_test:
        _, scores = eval_model(model, test_data, use_cuda=args.cuda, with_label=False)

        predict_file = os.path.join(output_dir, "predict.csv")
        dump_predictions(scores, labels, predict_file)
        logger.info("[new predictions saved.]")


if __name__ == "__main__":
    main()



from pathlib import Path

import pytest

from gobbli.test.util import MockModel
from gobbli.util import gobbli_version, read_metadata, write_metadata


def test_base_model_init(tmpdir):
    tmpdir_path = Path(tmpdir)
    MODEL_PARAMS = {"param": "a"}

    # Create model
    m = MockModel(data_dir=tmpdir_path, **MODEL_PARAMS)

    # Metadata path for the model should now exist
    assert m.metadata_path.exists()

    # So should the info path
    assert m.info_path.exists()

    # Info should be populated appropriately
    info = read_metadata(m.info_path)
    assert info["class"] == m.__class__.__name__
    assert info["gobbli_version"] == gobbli_version()

    # We shouldn't be able to create a new model in the same directory
    # with load_existing=False
    with pytest.raises(ValueError):
        MockModel(data_dir=tmpdir_path, load_existing=False, **MODEL_PARAMS)

    # We should be able to load the existing model, and it should
    # have the same param values without being passed explicitly
    m2 = MockModel(data_dir=tmpdir_path, load_existing=True)
    assert m2.params == MODEL_PARAMS


def test_base_model_init_warnings_errors(tmpdir):
    tmpdir_path = Path(tmpdir)

    # Create model to bootstrap the metadata/info files
    m = MockModel(data_dir=tmpdir_path, param="a")

    # Make sure we get a warning if user passes params that are ignored
    with pytest.warns(UserWarning):
        MockModel(data_dir=tmpdir_path, load_existing=True, param="b")

    # Make sure we get a warning if the gobbli version in the info file changes
    info = read_metadata(m.info_path)
    bad_version_info = info.copy()
    bad_version_info["gobbli_version"] = "not a real version"
    write_metadata(bad_version_info, m.info_path)
    with pytest.warns(UserWarning):
        MockModel(data_dir=tmpdir_path, load_existing=True)

    # Make sure we get an error if the class in the info file changes
    bad_class_info = info.copy()
    bad_class_info["class"] = "not a real model"
    write_metadata(bad_class_info, m.info_path)
    with pytest.raises(ValueError):
        MockModel(data_dir=tmpdir_path, load_existing=True)



import pandas as pd

from gobbli.dataset.imdb import IMDBDataset


def test_load_imdb(tmp_gobbli_dir):
    ds = IMDBDataset.load()

    X_train = ds.X_train()
    X_test = ds.X_test()

    y_train = ds.y_train()
    y_test = ds.y_test()

    assert len(X_train) == 25000
    assert len(y_train) == 25000
    assert len(X_test) == 25000
    assert len(y_test) == 25000

    assert len(pd.unique(y_train)) == 2
    assert len(pd.unique(y_test)) == 2

    # Ensure these objects pass validation
    ds.train_input()
    ds.predict_input()



# Copyright (c) Microsoft. All rights reserved.
import torch
import math
from torch.nn.functional import tanh, relu, prelu, leaky_relu, sigmoid, elu, selu

def linear(x):
    return x

def swish(x):
    return x * sigmoid(x)

def gelu(x):
    """ref:https://github.com/huggingface/pytorch-pretrained-BERT/blob/master/pytorch_pretrained_bert/modeling.py#L113
    """
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))

def activation(func_a):
    """Activation function wrapper
    """
    try:
        f = eval(func_a)
    except:
        f = linear
    return f



import re
import shutil
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score

import gobbli.io
from gobbli.docker import maybe_mount, run_container
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import EmbedMixin, PredictMixin, TrainMixin
from gobbli.util import (
    assert_in,
    assert_type,
    download_archive,
    escape_line_delimited_text,
    pred_prob_to_pred_label,
)

FASTTEXT_VECTOR_ARCHIVES = {
    "wiki-news-300d": "https://dl.fbaipublicfiles.com/fasttext/vectors-english/wiki-news-300d-1M.vec.zip",
    "wiki-news-300d-subword": "https://dl.fbaipublicfiles.com/fasttext/vectors-english/wiki-news-300d-1M-subword.vec.zip",
    "crawl-300d": "https://dl.fbaipublicfiles.com/fasttext/vectors-english/crawl-300d-2M.vec.zip",
    "crawl-300d-subword": "https://dl.fbaipublicfiles.com/fasttext/vectors-english/crawl-300d-2M-subword.zip",
    "wiki-crawl-300d": "https://dl.fbaipublicfiles.com/fasttext/vectors-crawl/cc.en.300.vec.gz",
    "wiki-aligned-300d": "https://dl.fbaipublicfiles.com/fasttext/vectors-aligned/wiki.en.align.vec",
}
"""
A mapping from pretrained vector names to archives.
See `the fastText docs <https://fasttext.cc/docs/en/english-vectors.html>`__ for information
about each set of vectors.  Note, some sets of vectors are very, very large.
"""

_DIM_REGEX = re.compile(r"([0-9]+)d")


def _parse_dim(model_name: str) -> int:
    """
    Parse the number of dimensions from a FastText model name.
    """
    match = _DIM_REGEX.search(model_name)
    if match is None:
        raise ValueError(
            f"Failed to parse number of dimensions from model name {model_name}"
        )

    return int(match.group(1))


def _fasttext_preprocess(text: str) -> str:
    """
    Preprocess text for the fasttext model.

    Lowercase text and escape newlines.  Removing or separately
    tokenizing punctuation is recommended, but there are too many different
    ways to do it, so we leave that up to the user.
    """
    return escape_line_delimited_text(text).lower()


@dataclass
class FastTextCheckpoint:
    path: Path

    @property
    def vectors(self) -> Path:
        """
        From a checkpoint, return the path to the text vectors.
        """
        return self.path.parent / f"{self.path.stem}.vec"

    @property
    def model(self) -> Path:
        """
        From a checkpoint, return the path to the binary model.
        """
        return self.path.parent / f"{self.path.stem}.bin"


class FastText(BaseModel, TrainMixin, PredictMixin, EmbedMixin):
    """
    Wrapper for Facebook's fastText model:
    https://github.com/facebookresearch/fastText

    Note: fastText benefits from some preprocessing steps:
    https://fasttext.cc/docs/en/supervised-tutorial.html#preprocessing-the-data

    gobbli will only lowercase and escape newlines in your input by default.
    If you want more sophisticated preprocessing for punctuation, stemming, etc,
    consider performing some preprocessing on your own beforehand.
    """

    _BUILD_PATH = Path(__file__).parent

    _TRAIN_INPUT_FILE = "train.txt"
    _VALID_INPUT_FILE = "valid.txt"
    _TEST_INPUT_FILE = "test.txt"

    _PREDICT_OUTPUT_FILE = "predict.txt"

    _EMBEDDING_INPUT_FILE = "input.txt"
    _EMBEDDING_OUTPUT_FILE = "embeddings.txt"

    _CHECKPOINT_BASE = "model"

    @property
    def image_tag(self) -> str:
        """
        Returns:
          The tag to use for the fastText image.
        """
        return "gobbli-fasttext"

    @property
    def weights_dir(self) -> Path:
        """
        Returns:
          The directory containing pretrained weights for this instance.
        """
        # Weights won't be used if we don't have a model to use
        if self.fasttext_model is None:
            return self.class_weights_dir
        return self.class_weights_dir / self.fasttext_model

    def init(self, params: Dict[str, Any]):
        """
        See :meth:`gobbli.model.base.BaseModel.init`.

        For more info on fastText parameter semantics, see
        `the docs <https://fasttext.cc/docs/en/options.html>`__.  The fastText
        `supervised tutorial <https://fasttext.cc/docs/en/supervised-tutorial.html>`__ has
        some more detailed explanation.

        fastText parameters:

        - ``word_ngrams`` (:obj:`int`): Max length of word n-grams.
        - ``lr`` (:obj:`float`): Learning rate.
        - ``dim`` (:obj:`int`): Dimension of learned vectors.
        - ``ws`` (:obj:`int`): Context window size.
        - ``autotune_duration`` (:obj:`int`): Duration in seconds to spend autotuning parameters.
          Any of the above parameters will not be autotuned if they are manually specified.
        - ``autotune_modelsize`` (:obj:`str`): Maximum size of autotuned model (ex "2M" for 2
          megabytes).  Any of the above parameters will not be autotuned if they are manually
          specified.
        - ``fasttext_model`` (:obj:`str`): Name of a pretrained fastText model to use.
          See :obj:`FASTTEXT_VECTOR_ARCHIVES` for a listing of available pretrained models.
        """
        self.word_ngrams = None
        self.lr = None
        self.ws = None
        self.fasttext_model = None
        # Default to dimensionality of the passed model, if any
        if "fasttext_model" in params:
            self.dim: Optional[int] = _parse_dim(params["fasttext_model"])
        else:
            self.dim = None
        self.autotune_duration = None
        self.autotune_modelsize = None

        for name, value in params.items():
            if name == "word_ngrams":
                assert_type(name, value, int)
                self.word_ngrams = value
            elif name == "lr":
                assert_type(name, value, float)
                self.lr = value
            elif name == "dim":
                assert_type(name, value, int)
                self.dim = value
            elif name == "ws":
                assert_type(name, value, int)
                self.ws = value
            elif name == "fasttext_model":
                assert_in(name, value, set(FASTTEXT_VECTOR_ARCHIVES.keys()))
                self.fasttext_model = value
            elif name == "autotune_duration":
                assert_type(name, value, int)
                self.autotune_duration = value
            elif name == "autotune_modelsize":
                assert_type(name, value, str)
                self.autotune_modelsize = value
            else:
                raise ValueError(f"Unknown param '{name}'")

        if (
            self.fasttext_model is not None
            and f"{self.dim}d" not in self.fasttext_model
        ):
            raise ValueError(
                "When using pretrained vectors, 'dim' must match the"
                f" dimensionality of the vectors; 'dim' value of {self.dim}"
                f" is incompatible with vectors {self.fasttext_model}."
            )

    def _build(self):
        # Download data if we need it and don't already have it
        if (
            self.fasttext_model is not None
            and not (self.weights_dir / self.fasttext_model).exists()
        ):
            with tempfile.TemporaryDirectory() as tmpdir:
                tmp_weights_dir = Path(tmpdir) / self.weights_dir.name
                tmp_weights_dir.mkdir()
                self.logger.info("Downloading pre-trained weights.")
                download_archive(
                    FASTTEXT_VECTOR_ARCHIVES[self.fasttext_model], tmp_weights_dir
                )
                shutil.move(tmp_weights_dir, self.weights_dir)
                self.logger.info("Weights downloaded.")

        # Build the custom docker image
        self.docker_client.images.build(
            path=str(FastText._BUILD_PATH),
            tag=self.image_tag,
            **self._base_docker_build_kwargs,
        )

    @staticmethod
    def _locate_checkpoint(weights_dir: Path) -> FastTextCheckpoint:
        """
        Locate a fastText checkpoint under the given directory,
        regardless of its filename.

        Args:
          weights_dir: The directory to search for a checkpoint (not recursive).

        Returns:
          A fastText checkpoint.
        """
        candidates = list(weights_dir.glob("*.vec"))
        if len(candidates) == 0:
            raise ValueError(f"No weights files found in '{weights_dir}'.")
        elif len(candidates) > 1:
            raise ValueError(
                f"Multiple weights files found in '{weights_dir}': {candidates}"
            )

        return FastTextCheckpoint(path=candidates[0].parent / candidates[0].stem)

    def _get_checkpoint(
        self, user_checkpoint: Optional[Path], context: ContainerTaskContext
    ) -> Tuple[Optional[FastTextCheckpoint], Optional[FastTextCheckpoint]]:
        """
        Determines, if any, the host checkpoint file and container checkpoint file
        using the user-requested checkpoint and the container context.

        Args:
          user_checkpoint: An optional checkpoint passed in by the user.  If the user doesn't
            pass one, use the default pretrained checkpoint, if any, or no checkpoint.
          context: The container context to create the checkpoint in.

        Returns:
          A 2-tuple: the host checkpoint (if any) and
            the container checkpoint (if any)
        """
        host_checkpoint = None  # type: Optional[FastTextCheckpoint]
        container_checkpoint = None  # type: Optional[FastTextCheckpoint]

        if self.fasttext_model is None and user_checkpoint is None:
            # No pretrained vectors
            return host_checkpoint, container_checkpoint

        elif self.fasttext_model is not None and user_checkpoint is None:
            host_checkpoint = FastText._locate_checkpoint(self.weights_dir)
            container_checkpoint = FastTextCheckpoint(
                BaseModel._CONTAINER_WEIGHTS_PATH / host_checkpoint.path.name
            )
            return host_checkpoint, container_checkpoint

        else:  # user_checkpoint is not None; user_checkpoint overrides pretrained model

            # This should never happen by the conditional checks above
            assert user_checkpoint is not None

            host_checkpoint = FastTextCheckpoint(user_checkpoint)
            container_checkpoint = FastTextCheckpoint(
                BaseModel._CONTAINER_WEIGHTS_PATH / host_checkpoint.path.name
            )

            return host_checkpoint, container_checkpoint

    def _write_input(
        self, X: List[str], y: Optional[List[str]], input_path: Path
    ) -> List[str]:
        """
        Write the given input and labels (if any) into the format expected by fastText.
        Make sure the given directory exists first.

        Returns:
          A sorted list of unique labels found in the dataset.
        """
        label_set = set()
        with open(input_path, "w") as f:
            if y is not None:
                for text, label in zip(X, y):
                    f.write(f"__label__{label} {_fasttext_preprocess(text)}\n")
                    label_set.add(label)
            elif y is None:
                for text in X:
                    f.write(f"{_fasttext_preprocess(text)}\n")

        return list(sorted(label_set))

    def _run_supervised(
        self,
        user_checkpoint: Optional[Path],
        container_input_path: Path,
        container_output_path: Path,
        context: ContainerTaskContext,
        num_epochs: int,
        autotune_validation_file_path: Optional[Path] = None,
        freeze_vectors: bool = False,
    ) -> Tuple[str, float]:
        """
        Run the fastText "supervised" command.  Used for both training and getting
        validation loss.

        Args:
          user_checkpoint: A checkpoint passed by the user
          container_input_path: Path to the input file in the container
          container_output_path: Path to the output checkpoint in the container
          context: Container task context.
        validation_file_path: Optional file to use for autotune validation when training.
          freeze_vectors: If true, use 0 learning rate; train solely for
            the purpose of calculating loss.

        Returns:
          A 2-tuple: container logs and loss.
        """
        host_checkpoint, container_checkpoint = self._get_checkpoint(
            user_checkpoint, context
        )

        cmd = (
            "supervised"
            f" -input {container_input_path}"
            f" -output {container_output_path}"
            f" -epoch {num_epochs}"
        )

        if autotune_validation_file_path is not None:
            cmd += f" -autotune-validation {autotune_validation_file_path}"

        lr = self.lr
        if freeze_vectors:
            lr = 0.0
        if lr is not None:
            cmd += f" -lr {lr}"

        for arg_name, attr in (
            ("wordNgrams", "word_ngrams"),
            ("dim", "dim"),
            ("ws", "ws"),
            ("autotune-duration", "autotune_duration"),
            ("autotune-modelsize", "autotune_modelsize"),
        ):
            attr_val = getattr(self, attr)
            if attr_val is not None:
                cmd += f" -{arg_name} {attr_val}"

        run_kwargs = self._base_docker_run_kwargs(context)

        if host_checkpoint is not None and container_checkpoint is not None:
            maybe_mount(
                run_kwargs["volumes"],
                host_checkpoint.vectors,
                container_checkpoint.vectors,
            )
            cmd += f" -pretrainedVectors {container_checkpoint.vectors}"

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        # Parse the training loss out of the console output
        last_loss_ndx = container_logs.rfind("avg.loss:")
        failed_parse_msg = (
            "Failed to parse loss information from fastText container logs."
            " Run with debug logging to"
            " see why this might have happened."
        )
        if last_loss_ndx == -1:
            raise ValueError(failed_parse_msg)

        # Skip over the word "avg.loss:" - next field in the output is "ETA:"
        loss_start_ndx = last_loss_ndx + len("avg.loss:")
        loss_end_ndx = container_logs.find("ETA:", loss_start_ndx)
        loss = float(container_logs[loss_start_ndx:loss_end_ndx].strip())
        return container_logs, loss

    def _run_predict_prob(
        self,
        user_checkpoint: Path,
        labels: List[str],
        container_input_path: Path,
        context: ContainerTaskContext,
    ) -> Tuple[str, pd.DataFrame]:
        """
        Run the fastText "predict-prob" command.  Used for obtaining
        label predicted probabilities on a dataset.

        Args:
          container_trained_model_path: Trained model passed by the user (.bin file)
          labels: Set of all labels to be used in prediction.
          container_input_path: Path to the input file in the container.
          context: Container task context.

        Returns:
          A 2-tuple: container logs and a dataframe of predicted probabilities.
        """
        host_checkpoint, container_checkpoint = self._get_checkpoint(
            user_checkpoint, context
        )

        if host_checkpoint is None or container_checkpoint is None:
            raise ValueError("A trained checkpoint is required to run prediction.")

        host_output_path = context.host_output_dir / FastText._PREDICT_OUTPUT_FILE
        container_output_path = (
            context.container_output_dir / FastText._PREDICT_OUTPUT_FILE
        )

        cmd = (
            "bash -c './fasttext predict-prob"
            f" {container_checkpoint.model}"
            f" {container_input_path}"
            f" {len(labels)}"
            f" >{container_output_path}'"
        )

        run_kwargs = self._base_docker_run_kwargs(context)
        # Override the entrypoint so we can use 'bash -c ...' above
        run_kwargs["entrypoint"] = ""
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint.model, container_checkpoint.model
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        # Parse the predicted probabilities out of the output file
        pred_prob_data = []
        with open(host_output_path, "r") as f:
            for line in f:
                tokens = line.split()
                row_data = {}
                for raw_label, prob in zip(tokens[0::2], tokens[1::2]):
                    # Strip the "__label__" prefix
                    label = raw_label[9:]
                    row_data[label] = float(prob)
                pred_prob_data.append(row_data)

        return (container_logs, pd.DataFrame(pred_prob_data))

    def _train(
        self, train_input: gobbli.io.TrainInput, context: ContainerTaskContext
    ) -> gobbli.io.TrainOutput:
        labels = self._write_input(
            train_input.X_train,
            train_input.y_train,
            context.host_input_dir / FastText._TRAIN_INPUT_FILE,
        )
        self._write_input(
            train_input.X_valid,
            train_input.y_valid,
            context.host_input_dir / FastText._VALID_INPUT_FILE,
        )

        container_validation_input_path = (
            context.container_input_dir / FastText._VALID_INPUT_FILE
        )
        train_logs, train_loss = self._run_supervised(
            train_input.checkpoint,
            context.container_input_dir / FastText._TRAIN_INPUT_FILE,
            context.container_output_dir / FastText._CHECKPOINT_BASE,
            context,
            train_input.num_train_epochs,
            autotune_validation_file_path=container_validation_input_path,
        )

        host_checkpoint_path = context.host_output_dir / f"{FastText._CHECKPOINT_BASE}"

        # Calculate validation accuracy on our own, since the CLI only provides
        # precision/recall
        predict_logs, pred_prob_df = self._run_predict_prob(
            host_checkpoint_path, labels, container_validation_input_path, context
        )
        pred_labels = pred_prob_to_pred_label(pred_prob_df)

        # Not ideal, but fastText doesn't provide a way to get validation loss;
        # Negate the validation accuracy instead
        valid_accuracy = accuracy_score(train_input.y_valid, pred_labels)
        valid_loss = -valid_accuracy

        return gobbli.io.TrainOutput(
            train_loss=train_loss,
            valid_loss=valid_loss,
            valid_accuracy=valid_accuracy,
            labels=labels,
            checkpoint=host_checkpoint_path,
            _console_output="\n".join((train_logs, predict_logs)),
        )

    def _predict(
        self, predict_input: gobbli.io.PredictInput, context: ContainerTaskContext
    ) -> gobbli.io.PredictOutput:
        host_input_path = context.host_input_dir / FastText._TEST_INPUT_FILE
        self._write_input(predict_input.X, None, host_input_path)
        container_input_path = context.to_container(host_input_path)

        if predict_input.checkpoint is None:
            raise ValueError("fastText requires a trained checkpoint for prediction.")

        predict_logs, pred_prob_df = self._run_predict_prob(
            predict_input.checkpoint,
            predict_input.labels,
            container_input_path,
            context,
        )

        return gobbli.io.PredictOutput(
            y_pred_proba=pred_prob_df, _console_output=predict_logs
        )

    def _embed(
        self, embed_input: gobbli.io.EmbedInput, context: ContainerTaskContext
    ) -> gobbli.io.EmbedOutput:
        # Check for null checkpoint here to give quick feedback to the user
        if embed_input.checkpoint is None:
            raise ValueError(
                "fastText requires a trained checkpoint to generate embeddings."
            )
        if embed_input.pooling == gobbli.io.EmbedPooling.NONE:
            raise ValueError(
                "fastText prints sentence vectors, so pooling is required."
            )

        host_input_path = context.host_input_dir / FastText._EMBEDDING_INPUT_FILE
        self._write_input(embed_input.X, None, host_input_path)
        container_input_path = context.to_container(host_input_path)

        host_checkpoint, container_checkpoint = self._get_checkpoint(
            embed_input.checkpoint, context
        )

        # We shouldn't get Nones here if the user didn't pass a null checkpoint, but
        # check anyway to satisfy mypy
        if host_checkpoint is None or container_checkpoint is None:
            raise ValueError(
                "fastText requires a trained checkpoint to generate embeddings."
            )

        host_output_path = context.host_output_dir / FastText._EMBEDDING_OUTPUT_FILE
        container_output_path = (
            context.container_output_dir / FastText._EMBEDDING_OUTPUT_FILE
        )

        cmd = (
            "bash -c './fasttext print-sentence-vectors"
            f" {container_checkpoint.model}"
            f" <{container_input_path}"
            f" >{container_output_path}'"
        )

        run_kwargs = self._base_docker_run_kwargs(context)
        # Override the entrypint so we can use 'bash -c ...' above
        run_kwargs["entrypoint"] = ""
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint.model, container_checkpoint.model
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        # Parse the embeddings out of the output file
        embeddings = np.loadtxt(host_output_path, comments=None, ndmin=2)

        return gobbli.io.EmbedOutput(
            X_embedded=embeddings, embed_tokens=None, _console_output=container_logs
        )



from gobbli.model.bert import BERT
from gobbli.model.fasttext import FastText
from gobbli.model.majority import MajorityClassifier
from gobbli.model.mtdnn import MTDNN
from gobbli.model.random import RandomEmbedder
from gobbli.model.sklearn import SKLearnClassifier, TfidfEmbedder
from gobbli.model.spacy import SpaCyModel
from gobbli.model.transformer import Transformer
from gobbli.model.use import USE

__all__ = [
    "BERT",
    "FastText",
    "MajorityClassifier",
    "MTDNN",
    "RandomEmbedder",
    "Transformer",
    "SKLearnClassifier",
    "SpaCyModel",
    "USE",
    "TfidfEmbedder",
]
